package com.galaxis.manatee.service;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.galaxis.manatee.constant.ChuanYunConstant;
import com.galaxis.manatee.dao.ChuanyunGroupProjectDao;
import com.galaxis.manatee.dao.ChuanyunSelfWorkHourDao;
import com.galaxis.manatee.dao.ChuanyunWorkHourDao;
import com.galaxis.manatee.entity.chuanyun.data.object.ChuanyunMemberHourDO;
import com.galaxis.manatee.entity.chuanyun.data.object.ChuanyunSelfWorkHourDO;
import com.galaxis.manatee.entity.chuanyun.data.object.ChuanyunUserCompanyDO;
import com.galaxis.manatee.entity.chuanyun.data.object.ChuanyunWorkHourDO;
import com.galaxis.manatee.entity.chuanyun.dto.ChuanyunSaveDTO;
import com.galaxis.manatee.entity.chuanyun.dto.Filter;
import com.galaxis.manatee.manager.ChuanYunManager;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 更新日志中所用到的服务
 *
 * @author zcj
 * @version 0.1
 * @date 2021/3/4 1:54 上午
 */
@Slf4j
@Service
public class LogService {
    private final ObjectMapper objectMapper = new ObjectMapper();
    private final ChuanyunWorkHourDao chuanyunWorkHourDao;
    private final ChuanyunSelfWorkHourDao chuanyunSelfWorkHourDao;
    private final ChuanYunManager chuanYunManager;
    private final ChuanyunGroupProjectDao chuanyunGroupProjectDao;

    public LogService(ChuanyunWorkHourDao chuanyunWorkHourDao, ChuanyunSelfWorkHourDao chuanyunSelfWorkHourDao, ChuanYunManager chuanYunManager, ChuanyunGroupProjectDao chuanyunGroupProjectDao) {
        this.chuanyunWorkHourDao = chuanyunWorkHourDao;
        this.chuanyunSelfWorkHourDao = chuanyunSelfWorkHourDao;
        this.chuanYunManager = chuanYunManager;
        this.chuanyunGroupProjectDao = chuanyunGroupProjectDao;
    }

    /**
     * 员工单日、各个项目工时标准化,并保存
     *
     * @param isRecent              是否只标准化最近一个月的数据
     * @param chuanyunUserCompanyDO 员工信息
     */
    public void standardHour(ChuanyunUserCompanyDO chuanyunUserCompanyDO, boolean isRecent) {
        if (!StringUtils.isEmpty(chuanyunUserCompanyDO.getUserId())) {
            //从日志原始数据中获取每人每个项目每天工时。
            //由isRecent判断是否只获取最近30天的数据
            List<ChuanyunWorkHourDO> selfList = isRecent ? chuanyunWorkHourDao.getSelfWorkHourRecently(chuanyunUserCompanyDO.getUserId()) : chuanyunWorkHourDao.getSelfWorkHour(chuanyunUserCompanyDO.getUserId());
            //逐一将原始工时数据标准化后，更新到标准化工时表和氚云中
            selfList.forEach(self -> {
                try {
                    //新建标准化工时对象，并将原始数据更新到此对象中，并计算标准化之后的工时
                    ChuanyunSelfWorkHourDO chuanyunSelfWorkHourDO=this.fromChuanyunWorkHourDO(self);
                    //查询系统中是否有重复的标准化工时数据，如果有重复的人、天、项目数据，将原数据删除后更新为最新的数据
                    try{
                        ChuanyunSelfWorkHourDO tmpChuanyunSelfWorkHourDO = chuanyunSelfWorkHourDao.findByUserIdAndAndProjectIdAndAndDayLogDateAndProjectType(chuanyunUserCompanyDO.getUserId(), ChuanYunConstant.PRODUCTION_PROJECT.equals(self.getProjectType()) ? ChuanYunConstant.PRODUCTION_PROJECT_ID : self.getProjectId(), self.getDayLogDate(), self.getProjectType());
                        if (tmpChuanyunSelfWorkHourDO != null) {
                            chuanyunSelfWorkHourDao.delete(tmpChuanyunSelfWorkHourDO);
                        }
                        chuanyunSelfWorkHourDao.save(chuanyunSelfWorkHourDO);
                    }catch (Exception e){
                        log.error(e.getMessage());
                    }
                } catch (Exception e) {
                    log.error(e.getMessage());
                    log.error("" + self);
                    e.printStackTrace();
                }
            });
        }
    }

    /**
     * 核对日志原始数据与每日工时标准化数据差异
     *
     * @param userId 人员对象ID
     */
    public void checkDayWorkHour(String userId) {
        long start=System.currentTimeMillis();
        log.info("开始核对日志原始工时与标准化工时");
        //根据日志原始数据获取单个人的所有工时记录
        List<ChuanyunWorkHourDO> chuanyunWorkHourList = chuanyunWorkHourDao.getSelfWorkHour(userId);
        //根据标准化工时获取单个人的标准化工时记录
        List<ChuanyunSelfWorkHourDO> chuanyunSelfWorkHourList = chuanyunSelfWorkHourDao.findByUserId(userId);
        //删除多余每日工时
        compareDayWorkHour(chuanyunWorkHourList, chuanyunSelfWorkHourList);
        log.info("核对耗时"+(System.currentTimeMillis()-start)+"毫秒");
    }

    /**
     * 删除多余每日工时
     *
     * @param chuanyunWorkHourList      日志原始工时
     * @param chuanyunSelfWorkHourList  标准化工时列表
     */
    private void compareDayWorkHour(List<ChuanyunWorkHourDO> chuanyunWorkHourList, List<ChuanyunSelfWorkHourDO> chuanyunSelfWorkHourList) {
        Map<String, Integer> map = new HashMap<>(64);
        //判断工时是否有撤回
        if (chuanyunWorkHourList.size() != chuanyunSelfWorkHourList.size()) {
            for (ChuanyunWorkHourDO chuanyunWorkHourDO : chuanyunWorkHourList) {
                map.put(chuanyunWorkHourDO.getUserId() + chuanyunWorkHourDO.getProjectId() + chuanyunWorkHourDO.getDayLogDate(), 1);
            }
            //比较是否有撤回工时
            for (ChuanyunSelfWorkHourDO chuanyunSelfWorkHourDO : chuanyunSelfWorkHourList) {
                Integer count = map.get(chuanyunSelfWorkHourDO.getUserId() + chuanyunSelfWorkHourDO.getProjectId() + chuanyunSelfWorkHourDO.getDayLogDate());
                if (count == null) {
                    try {
                        //删除数据库日工时
                        chuanyunSelfWorkHourDao.delete(chuanyunSelfWorkHourDO);
                        //删除氚云日工时
                        List<String> matchers = new ArrayList<>();
                        matchers.add("F0000001_2," + chuanyunSelfWorkHourDO.getProjectId());
                        matchers.add("F0000002_2," + chuanyunSelfWorkHourDO.getUserId());
                        matchers.add("F0000003_2," + chuanyunSelfWorkHourDO.getDayLogDate());
                        var filter = Filter.instance(0, 1, true, "And", matchers);
                        var chuanyunFindAllResponse = chuanYunManager.findAll(ChuanyunMemberHourDO.SCHEMA_CODE, filter);
                        ChuanyunSaveDTO chuanyunSaveDTO;
                        if (chuanyunFindAllResponse.getReturnData() != null) {
                            List<ChuanyunMemberHourDO> result = objectMapper.convertValue(chuanyunFindAllResponse.getReturnData().getBizObjectArray(), new TypeReference<>() {
                            });
                            chuanyunSaveDTO = chuanYunManager.delete(ChuanyunMemberHourDO.SCHEMA_CODE, result.get(0).getObjectId());
                        }
                        log.info(chuanyunSelfWorkHourDO.getUserId() + "删除" + chuanyunSelfWorkHourDO.getDayLogDate() + "位于" + chuanyunSelfWorkHourDO.getProjectId() + "工时");
                    } catch (Exception e) {
                        e.printStackTrace();
                        log.error(e.getMessage());
                    }
                }
            }
        }
    }

    /**
     * 根据日志原始数据统计的每日每人工时获取工时标准化后的标准化对象
     * @param chuanyunWorkHourDO    日志原始统计数据
     * @return  工时标准化对象
     */
    private ChuanyunSelfWorkHourDO fromChuanyunWorkHourDO(ChuanyunWorkHourDO chuanyunWorkHourDO) throws Exception {
        //由于日志原始工时可能存在项目类型不准确的情况，需要根据项目ID获取真实的项目类型
        String projectType;
        try {
            var chuanyunGroupProject = chuanyunGroupProjectDao.findByObjectId(chuanyunWorkHourDO.getProjectId());
            if (chuanyunGroupProject == null) {
                throw new Exception("没有找到对应的项目");
            }
            //如果找到相应的项目，则使用项目实际的项目类型
            projectType=chuanyunGroupProject.getProjectType();
        } catch (Exception e) {
            //没有找到对应项目，使用原项目类型
            projectType=chuanyunWorkHourDO.getProjectType();
        }
        ChuanyunSelfWorkHourDO chuanyunSelfWorkHourDO = ChuanyunSelfWorkHourDO.fromChuanyunWorkHourDO(chuanyunWorkHourDO,projectType);
        //获取每人每天总工时
        ChuanyunWorkHourDO oneDayTotalHour = chuanyunWorkHourDao.getDayWorkHour(chuanyunWorkHourDO.getUserId(), chuanyunWorkHourDO.getDayLogDate());
        //工时标准化
        if (oneDayTotalHour != null) {
            BigDecimal standard = new BigDecimal(8);
            if (oneDayTotalHour.getHoursWorked().compareTo(standard) <= 0) {
                chuanyunSelfWorkHourDO.setStandardWorkHour(chuanyunWorkHourDO.getHoursWorked());
            } else {
                BigDecimal personDays = (chuanyunWorkHourDO.getHoursWorked().divide(oneDayTotalHour.getHoursWorked(), 2, RoundingMode.HALF_UP)).multiply(standard);
                chuanyunSelfWorkHourDO.setStandardWorkHour(personDays);
            }
        }else{
            throw new Exception("获取每日总工时失败");
        }
        return chuanyunSelfWorkHourDO;
    }
}
